import numpy as np

from typing import Tuple, Optional

from ..pylena_cxx import morpho as cxx
from ..utils import check_numpy_array


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
def watershed(
    img: np.ndarray,
    connectivity: int,
    waterline: bool = True,
    markers: Optional[np.ndarray] = None,
) -> Tuple[np.ndarray, int]:
    """
    Performs a Watershed segmentation using the Meyer's Watershed algorithm :cite:p:`meyer.91.rfia` on an image ``img``
    using the neighborhood relationship ``connectivity``.

    Args
    ----
    img: 2-D array (dtype=uint8)
        The input image
    connectivity: int
        The connectivity used by the watershed. Should be 4 (for 4-connectivity) or 8 (for 8-connectivity)
    waterline: bool
        Boolean to either display the watershed lines or
    markers: Optional[np.ndarray]
        Optional image (dtype=int16) representing the input markers. In this case, the watershed is computed from
        markers and not from local minima and it returns the watershed segmentation from this markers with watershed
        lines (``waterline`` argument is ignored in this case)

    Returns
    -------
    Tuple[np.ndarray, int]
        The resulting segmentation

    Raises
    ------
    ValueError
        If the connectivity is invalid or the input image is not & 2-D array with values encoded as uint8 value
    """
    if connectivity not in (4, 8):
        raise ValueError(f"Input connectivity should be 4 or 8 (Got {connectivity}")

    if markers is None:
        return cxx.watershed(img, connectivity, waterline)
    else:
        if markers.shape != img.shape or markers.dtype != np.int16:
            raise ValueError(
                f"Markers image should be an image with the same shape as the input image with value encoded in int16 (Got markers shape: {markers.ndim}, dtype: {markers.dtype})"
            )
        return cxx.watershed_from_markers(img, markers, connectivity)



import numpy as np

import pytest

from pylena import check_numpy_array, check_type


@check_type(ind=0, wanted_types=[float, int, np.ndarray])
def _func_check_type(a):
    return a + 1


@check_type(ind=1)
def _func2_check_type(a):
    return "error"


@check_numpy_array(ind=0, dtypes=[np.uint8, np.int16], ndims=[2, 3])
def _func_check_numpy_array(tab):
    return


@check_numpy_array(ind=1, dtypes=[np.uint8, np.uint16], ndims=[2])
def _func2_check_numpy_array(tab):
    return


def test_check_type():
    _func_check_type(float(1))
    _func_check_type(10)
    _func_check_type(np.ones((5,)))
    with pytest.raises(ValueError):
        _func_check_type((1, 2))
    with pytest.raises(ValueError):
        _func2_check_type(10)
    assert True


def test_check_numpy_array():
    _func_check_numpy_array(np.zeros((3, 3), dtype=np.uint8))
    _func_check_numpy_array(np.zeros((3, 3), dtype=np.int16))
    _func_check_numpy_array(np.zeros((3, 3, 3), dtype=np.uint8))
    _func_check_numpy_array(np.zeros((3, 3, 3), dtype=np.int16))
    with pytest.raises(ValueError):
        _func_check_numpy_array(np.zeros((3, 3), dtype=np.uint16))
    with pytest.raises(ValueError):
        _func_check_numpy_array(np.zeros((3, 3, 3, 3), dtype=np.uint8))
    with pytest.raises(ValueError):
        _func_check_numpy_array(np.zeros((3,), dtype=np.int16))
    with pytest.raises(ValueError):
        _func2_check_numpy_array(np.zeros((3, 3), dtype=np.uint8))
    with pytest.raises(ValueError):
        _func_check_numpy_array("tata")
    assert True



"""
Base
"""

__all__ = ["check_type", "check_numpy_array", "morpho", "scribo", "__version__"]

from . import morpho as morpho
from . import scribo as scribo
from .utils import *
from ._version import __version__



"""
Mathematical morphology
"""

from .se import make_structuring_element_2d
from .soperations import *
from .component_tree import maxtree, tos, ComponentTree
from .watershed import watershed

__all__ = [
    "make_structuring_element_2d",
    "erosion",
    "dilation",
    "opening",
    "closing",
    "gradient",
    "maxtree",
    "ComponentTree",
    "tos",
    "watershed",
]



import functools

import numpy as np


def check_type(func=None, ind=0, wanted_types=[]):
    """
    Decorator to check the type of an argument

    Args
    ----
    func
        The function to be decorated
    ind
        The index of the argument to be checked in the function
    wanted_types
        The accepted type for the argument

    Raise
    -----
    ValueError
        If `ind` is superior or equal to the number of argument or if the type is not accepted by the function

    Example
    -------
    .. code-block:: python

        @check_type(ind=0, wanted_types=[str])
        def foo(name):
            return "My name is {}".format(name)

    """

    def decorator(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            if ind >= len(args):
                raise ValueError(
                    "Missing argument in the function `{}`".format(func.__name__)
                )
            input_type = type(args[ind])
            ok = False
            for wanted_type in wanted_types:
                if issubclass(input_type, wanted_type):
                    ok = True
            if not ok:
                raise ValueError(
                    "Function `{}`: argument at index {} should be of type `{}` (Got `{}`)".format(
                        func.__name__,
                        ind,
                        [wanted_type.__name__ for wanted_type in wanted_types],
                        input_type.__name__,
                    )
                )
            return func(*args, **kwargs)

        return wrap

    if func is None:
        return decorator
    return decorator(func)  # coverage: disable


def check_numpy_array(func=None, ind=0, dtypes=[], ndims=[]):
    """
    Decorator to check that the input numpy array dtype is accepted by a
    function

    Args
    ----
    func
        The function to be decorated
    ind
        The index of the argument in the function
    dtypes
        The list of accepted dtypes
    ndims
        The list of accepted number of dimensions

    Raise
    -----
    ValueError
        If `ind` is superior or equal to the number of argument, the
        argument pointed by `ind` is not a numpy array, if the dtype is not accepted
        or if the number of dimensions is incorrect.

    Example
    -------

    .. code-block:: python

        @check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
        def transpose_uint8(a):
            return a.T
    """

    def decorator(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            if ind >= len(args):
                raise ValueError(
                    "Missing argument in the function `{}`".format(func.__name__)
                )
            if not issubclass(type(args[ind]), np.ndarray):
                raise ValueError(
                    "Function `{}`: argument at index {} should be a `ndarray` (Got `{}`)".format(
                        func.__name__, ind, type(args[ind]).__name__
                    )
                )
            if not args[ind].dtype in dtypes:
                raise ValueError(
                    "Function `{}`: argument at index {} has invalid numpy dtype (Got {}, expected in {})".format(
                        func.__name__,
                        ind,
                        args[ind].dtype,
                        [d.__name__ for d in dtypes],
                    )
                )
            if not args[ind].ndim in ndims:
                raise ValueError(
                    "Function `{}`: argument at index {} has invalid number of dimension (Got {}, expected in {})".format(
                        func.__name__, ind, args[ind].ndim, ndims
                    )
                )
            return func(*args, **kwargs)

        return wrap

    if func is None:
        return decorator
    return decorator(func)  # coverage: disable



"""
Structuring element
"""

from ..pylena_cxx.morpho import structuring_element_2d

import numpy as np

from warnings import warn


def make_structuring_element_2d(kind, *args):
    """
    Compute a predifined structuring element. The morphological operations for
    these structuring elements have been optimized for their use.

    Three kinds of structuring element can be computed using this method:

    * 2-D rectangle
    * Disc
    * 2-D periodic line

    For these structuring elements, argument taken by the function are different.

    * 2-D rectangle

    The supplementary arguments are the width and the height

    >>> se = make_structuring_element_2d("rect", width, height)

    * Disc

    The supplementary argument is the radius of the disc

    >>> se = make_structuring_element_2d("disc", radius)

    * 2-D periodic line

    The supplementary argument are the period (represented by a tuple of int of size 2 corresponding to a point)
    and the half number of pixel in the line

    >>> se = make_structuring_element_2d("periodic_line", (2, 5), 10)

    Args
    ----
    kind:
        The structuring element to be computed.
    args:
        The parameters for the computation of a structuring element

    Return
    ------
    structuring_element_2d
        A structuring element

    Raise
    -----
    ValueError
        If the `kind` argument is incorrect or if the argument for the structuring element computation is incorrect.

    Example
    -------

    The following code sample show how to compute a rectangle structuring element

    >>> se = make_structuring_element_2d("rect", 10, 5) # width=10 and height=5

    """
    if kind == "disc":
        if len(args) < 1:
            raise ValueError("Disc needs one argument (the radius)")
        if len(args) > 1:
            warn("The arguments after the radius will not be taken into account")
        return structuring_element_2d(float(args[0]))
    elif kind == "rect":
        if len(args) < 2:
            raise ValueError("Rect needs two arguments (the width and the height)")
        if len(args) > 2:
            warn(
                "The arguments after the width and the height will not be taken into account"
            )
        return structuring_element_2d(int(args[0]), int(args[1]))
    elif kind == "periodic_line":
        if len(args) < 2:
            raise ValueError("Periodic line: invalid number of argument")
        if (
            not isinstance(args[0], tuple)
            or len(args[0]) != 2
            or not isinstance(args[0][0], int)
            or not isinstance(args[0][1], int)
        ):
            raise ValueError(
                "First argument for periodic line is a tuple of int of size 2"
            )
        if len(args) > 2:
            warn("The arguments after the arguments will not be taken into account")
        return structuring_element_2d(args[0], int(args[1]))
    else:
        raise ValueError("Structuring element `{}` unknown".format(kind))



"""
Line detector
"""
from typing import Tuple, List, Union
from ..utils import check_numpy_array

from ..pylena_cxx import scribo as cxx
from ..pylena_cxx.scribo import e_segdet_preprocess, e_segdet_process_traversal_mode, e_segdet_process_tracking, e_segdet_process_extraction
from ..pylena_cxx.scribo import SegDetParams
from ..pylena_cxx.scribo import LSuperposition, VSegment

import numpy as np


def get_params(kwargs) -> Tuple[int, SegDetParams]:
    """
    Get the parameters for the line detection from the input arguments

    Return
    ------
    Tuple[int, SegDetParams]
        The minimum length of a line and the parameters for the line detection
    """
    min_len: int = kwargs.get("min_len", 10)

    params: SegDetParams = SegDetParams()
    params.preprocess = kwargs.get('preprocess', e_segdet_preprocess.NONE)
    params.tracker = kwargs.get('tracker', e_segdet_process_tracking.KALMAN)
    params.traversal_mode = kwargs.get('traversal_mode', e_segdet_process_traversal_mode.HORIZONTAL_VERTICAL)
    params.extraction_type = kwargs.get('extraction_type', e_segdet_process_extraction.BINARY)
    params.negate_image = kwargs.get('negate_image', False)
    params.dyn = kwargs.get('dyn', 0.6)
    params.size_mask = kwargs.get('size_mask', 11)
    params.double_exponential_alpha = kwargs.get('double_exponential_alpha', 0.6)
    params.simple_moving_average_memory = kwargs.get('simple_moving_average_memory', 30.0)
    params.exponential_moving_average_memory = kwargs.get('exponential_moving_average_memory', 16.0)
    params.one_euro_beta = kwargs.get('one_euro_beta', 0.007)
    params.one_euro_mincutoff = kwargs.get('one_euro_mincutoff', 1.0)
    params.one_euro_dcutoff = kwargs.get('one_euro_dcutoff', 1.0)
    params.bucket_size = kwargs.get('bucket_size', 32)
    params.nb_values_to_keep = kwargs.get('nb_values_to_keep', 30)
    params.discontinuity_relative = kwargs.get('discontinuity_relative', 0)
    params.discontinuity_absolute = kwargs.get('discontinuity_absolute', 0)
    params.minimum_for_fusion = kwargs.get('minimum_for_fusion', 15)
    params.default_sigma_position = kwargs.get('default_sigma_position', 2)
    params.default_sigma_thickness = kwargs.get('default_sigma_thickness', 2)
    params.default_sigma_luminosity = kwargs.get('default_sigma_luminosity', 57)
    params.min_nb_values_sigma = kwargs.get('min_nb_values_sigma', 10)
    params.sigma_pos_min = kwargs.get('sigma_pos_min', 1.0)
    params.sigma_thickness_min = kwargs.get('sigma_thickness_min', 0.64)
    params.sigma_luminosity_min = kwargs.get('sigma_luminosity_min', 13.0)
    params.gradient_threshold = kwargs.get('gradient_threshold', 30)
    params.llumi = kwargs.get('llumi', 225)
    params.blumi = kwargs.get('blumi', 225)
    params.ratio_lum = kwargs.get('ratio_lum', 1.0)
    params.max_thickness = kwargs.get('max_thickness', 100)
    params.threshold_intersection = kwargs.get('threshold_intersection', 0.8)
    params.remove_duplicates = kwargs.get('remove_duplicates', True)

    ret: Tuple[int, SegDetParams] = (min_len, params)

    return ret


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
def line_detector(
    img: np.ndarray,
    mode: str = "full",
    verbose: bool = False,
    **kwargs
) -> Union[List[VSegment], Tuple[np.ndarray, List[LSuperposition]], Tuple[np.ndarray, List[LSuperposition], List[VSegment]]]:
    """
    Perform a line detection in a greyscale document image from :cite:p:`bernet.23.icdar`

    Args
    ----
    img: 2-D array (dtype=uint8)
        The image to be processed
    mode: str
        The mode of the line detection. Should be one of 'full', 'pixel' or 'vector'
    verbose: bool
        Whether to print information about the line detection
    **kwargs: Additional keyword arguments.

    Keyword Arguments:
        min_len (int, optional): The minimum length of a line.
            Default is 10.
        preprocess (e_segdet_preprocess, optional): Preprocess applied.
            Default is e_segdet_preprocess.NONE.
        tracker (e_segdet_process_tracking, optional): Tracker used.
            Default is e_segdet_process_tracking.KALMAN.
        traversal_mode (e_segdet_process_traversal_mode, optional): Traversal performed.
            Default is e_segdet_process_traversal_mode.HORIZONTAL_VERTICAL.
        extraction_type (e_segdet_process_extraction, optional): Extraction type for observations.
            Default is e_segdet_process_extraction.BINARY.
        negate_image (bool, optional): Specify if the image has to be reversed before processing.
            Default is False.
        dyn (float, optional): Dynamic when Black-Top-Hat preprocess is applied.
            Default is 0.6.
        size_mask (int, optional): Filter size when Black-Top-Hat preprocess is applied.
            Default is 11.
        double_exponential_alpha (float, optional): Alpha used in double exponential tracker if chosen.
            Default is 0.6.
        simple_moving_average_memory (float, optional): Memory used in simple moving average tracker if chosen.
            Default is 30.0.
        exponential_moving_average_memory (float, optional): Memory used in exponential moving average tracker if chosen.
            Default is 16.0.
        one_euro_beta (float, optional): Beta used in one euro tracker if chosen.
            Default is 0.007.
        one_euro_mincutoff (float, optional): Min cutoff used in one euro tracker if chosen.
            Default is 1.0.
        one_euro_dcutoff (float, optional): Dcutoff used in one euro tracker if chosen.
            Default is 1.0.
        bucket_size (int, optional): Bucket size during traversal.
            Default is 32.
        nb_values_to_keep (int, optional): Memory of tracker to compute variances for the matching.
            Default is 30.
        discontinuity_relative (int, optional): Percentage. Discontinuity = discontinuity_absolute +
            discontinuity_relative * current_segment_size.
            Default is 0.
        discontinuity_absolute (int, optional): Discontinuity = discontinuity_absolute +
            discontinuity_relative * current_segment_size.
            Default is 0.
        minimum_for_fusion (int, optional): Threshold to merge trackers following the same observation.
            Default is 15.
        default_sigma_position (int, optional): Position default variance value.
            Default is 2.
        default_sigma_thickness (int, optional): Thickness default variance value.
            Default is 2.
        default_sigma_luminosity (int, optional): Luminosity default variance value.
            Default is 57.
        min_nb_values_sigma (int, optional): Threshold to compute variance and not use default values.
            Default is 10.
        sigma_pos_min (float, optional): Minimum position variance value.
            Default is 1.0.
        sigma_thickness_min (float, optional): Minimum thickness variance value.
            Default is 0.64.
        sigma_luminosity_min (float, optional): Minimum luminosity variance value.
            Default is 13.0.
        gradient_threshold (int, optional): Gradient threshold when gradient preprocess is applied.
            Default is 30.
        llumi (int, optional): First threshold for observation ternary extraction.
            Default is 225.
        blumi (int, optional): Second threshold for observation ternary extraction.
            Default is 225.
        ratio_lum (float, optional): Ratio of kept luminosity in observation extraction.
            Default is 1.0.
        max_thickness (int, optional): Max allowed (vertical|horizontal) thickness of segment to detect.
            Default is 100.
        threshold_intersection (float, optional): Threshold for duplication removal.
            Default is 0.8.
        remove_duplicates (bool, optional): Specify if duplication removal has to be computed.
            Default is True.

    Return
    ------
    List[VSegment] or Tuple[np.ndarray, List[LSuperposition]] or Tuple[np.ndarray, List[LSuperposition], List[VSegment]]
        The detected lines in the image or the label map, the superpositions and the segments of the detected lines depending on the mode of the line detection (see `mode`).

    Example
    -------
    >>> from pylena.scribo import line_detector
    >>> img = ... # Get an image
    >>> img_label, superpositions, lines = pln.scribo.line_detector(img, "full", min_len=100, blumi=110, llumi=110, discontinuity_relative=5)
    """
    min_len: int
    params: SegDetParams

    min_len, params = get_params(kwargs)
    if verbose:
        print("min_len = {}".format(min_len))
        print("params = {}".format(params))

    if mode == "full":
        return cxx.line_detector_full(img, min_len, params)
    elif mode == "pixel":
        return cxx.line_detector_pixel(img, min_len, params)
    elif mode == "vector":
        return cxx.line_detector_vector(img, min_len, params)
    else:
        raise ValueError(
            "Invalid mode. Should be one of 'full', 'pixel' or 'vector'")



"""
Document image operations.
"""

from .line_detector import line_detector
from .line_detector import e_segdet_preprocess, e_segdet_process_traversal_mode, e_segdet_process_tracking, e_segdet_process_extraction
from .line_detector import SegDetParams
from .line_detector import LSuperposition, VSegment

__all__ = [
    "line_detector",

    "e_segdet_preprocess",
    "e_segdet_process_traversal_mode",
    "e_segdet_process_tracking",
    "e_segdet_process_extraction",

    "LSuperposition",
    "VSegment",
]



from typing import Callable, Any, Tuple
from ..pylena_cxx import morpho as cxx
from ..utils import check_numpy_array
from dataclasses import dataclass

import numpy as np
from numba import jit


@jit(nopython=True)
def _compute_area(parent: np.ndarray, nodemap: np.ndarray):
    n = len(parent)
    A = np.zeros(n, dtype="int")
    for v in np.nditer(nodemap):
        A[v] += 1
    for i in range(n - 1, 0, -1):
        A[parent[i]] += A[i]
    return A


@jit(nopython=True)
def _filter(parent: np.ndarray, predicate: np.ndarray, values: np.ndarray, inplace):
    if not inplace:
        values = values.copy()
    n = len(parent)
    for i in range(1, n):
        if not predicate[i]:
            values[i] = values[parent[i]]
    return values


@jit(nopython=True)
def _compute_depth(parent: np.ndarray):
    n = len(parent)
    A = np.empty(n, dtype="int")
    A[0] = 0
    for i in range(1, n):
        A[i] = A[parent[i]] + 1
    return A


# @jit(nopython=False)
def _compute_attribute(
    parent: np.ndarray, nodemap: np.ndarray, fun: Callable, init_value
):
    n = len(parent)
    A = np.empty(n, dtype=object)
    A.fill(init_value)
    it = np.nditer(nodemap, ["multi_index"])
    for v in it:
        A[v] = fun(A[v], it.multi_index)
    for i in range(n - 1, 0, -1):
        A[parent[i]] = fun(A[parent[i]], A[i])
    return A


@dataclass
class ComponentTree:
    """Tree representation of the inclusion of some connected components of an image (max-tree, min-tree, or tree of shapes)"""

    parent: np.ndarray
    values: np.ndarray
    nodemap: np.ndarray

    def compute_area(self) -> np.ndarray:
        """Compute the area attribute map of an image.

        Returns
        -------
        np.ndarray
            A mapping :math:`n \\to \\text{area}` which return the area of a node :math:`n`.

        Example
        -------
        >>> t = pln.morpho.tos(img) # img is a (10, 10) image
        >>> area = t.compute_area()
        >>> area[0] # 0 is the index of the root of the tree
        100
        """
        return _compute_area(self.parent, self.nodemap)

    def compute_depth(self) -> np.ndarray:
        """Compute the depth attribute map of an image.

        Returns
        -------
        np.ndarray
            A mapping :math:`n \\to \\text{depth}` which return the depth of a node :math:`n`.

        Example
        -------
        >>> t = pln.morpho.tos(img)
        >>> depth = t.compute_depth()
        >>> depth[0] # 0 is the root, so its depth is 0
        0
        """
        return _compute_depth(self.parent)

    def compute_attribute(self, fun: Callable, init_value: Any) -> np.ndarray:
        """Compute an attribute using the callable object ``fun``

        Args
        ----
        fun: Callable
            A function ``fun(cur, arg)`` where ``cur`` is the value of the attribute of the current node and ``arg`` is the value taken for the current node.
        init_value: Any
            The value used to initialize the attribute

        Returns
        -------
        np.ndarray
            A mapping :math:`n \\to \\text{attr}` which return the attribute computed by ``fun`` for a node :math:`n`.

        Example
        -------
        >>> t = pln.morpho.tos(img) # img is a (10, 10) image
        >>> def area_acc(cur : int, arg):
        ...    if isinstance(arg, tuple): # Arg is a point of the nodemap
        ...        return cur + 1
        ...    return cur + arg           # Arg is the value of the mapping (here a child)
        >>> area = t.compute_attribute(area_acc, 0) # area_acc is an accumulator function to compute the area
        >>> area[0]
        100
        """
        return _compute_attribute(self.parent, self.nodemap, fun, init_value)

    def filter(
        self, predicate: np.ndarray, values: np.ndarray = None, inplace=True
    ) -> np.ndarray:
        """Filter (direct filtering rule) the tree based on a boolean predicate for each node.
        If **inplace**, the values array is modified inplace (a copy is returned otherwise).


        Args
        ----
        predicate: np.ndarray
            A boolean array that tells if a node as to be preserved or not.
        values: np.ndarray, optional
            The node values to filter. Defaults to ``self.values``.
        inplace: bool, optional
            Modify the values array or return a copy. Defaults to ``True``.

        Returns
        -------
        np.ndarray
            The filtered value array.

        Example
        -------
        >>> t = pln.morpho.tos(img)
        >>> area = t.compute_area()
        >>> t.filter(area >= 100) # Remove all the nodes whose area is smaller than 100
        """
        if values is None:
            values = self.values
        return _filter(self.parent, predicate, values, inplace)

    def reconstruct(self, values: np.ndarray = None) -> np.ndarray:
        """Reconstruct an image from the values array.

        Args
        ----
        values: np.ndarray, optional
            Values of the nodes. Defaults is self.values.

        Returns
        -------
        np.ndarray
            The reconstructed image

        Example
        -------
        >>> t = pln.morpho.tos(img)
        >>> area = t.compute_area()
        >>> new_values = t.filter(area >= 100, inplace=False)
        >>> rec = t.reconstruct(new_values)
        """
        if values is None:
            values = self.values
        return values[self.nodemap]


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
def maxtree(img: np.ndarray, connectivity: int) -> ComponentTree:
    """Compute the max-tree :cite:p:`salembier.98.tip` of a 2D 8-bit image.

    Args
    ----
    img: np.ndarray
        The input image
    connectivity: int
        Input connectivity used to compute the maxt-tree (4 for 4-neighborhood or 8 for 8-neighborhood)

    Raises
    ------
    ValueError
        If the input connectivity is incorrect

    Returns
    -------
    ComponentTree
        The computed max-tree

    Example
    -------
    >>> import pylena as pln
    >>> t = pln.morpho.maxtree(img, 4) # Compute the maxtree of an image using 4 connectivity.
    """
    if connectivity not in [4, 8]:
        raise ValueError(f"Connectivity should be 4 or 8 (not {connectivity})")
    tree, nodemap = cxx.maxtree(img, connectivity)
    return ComponentTree(tree.parent, tree.values, nodemap)


def add_border_median(f: np.array):
    m = np.median(np.concatenate([f[0, :], f[-1, :], f[:, 0], f[:, -1]]))
    return np.pad(f, [(1, 1), (1, 1)], mode="constant", constant_values=m)


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
def tos(
    img: np.ndarray,
    root: Tuple = None,
    padding: str = None,
    subsampling: str = "original",
) -> ComponentTree:
    """Compute the tree of shapes :cite:p:`monasse.00.tip` of a 2D image in linear complexity :cite:p:`carlinet.18.icip`

    Args
    ----
    input: np.ndarray
        An 2D image with values encoded as 8 bits unsigned integer.
    root: Tuple, optional
        The root point of the tree (in the form ``(row, column)``) used as a starting point in the propagation. Defaults to None.
    padding: str, optional
        Add an extra border if not None. Defaults to None.
        Options available:

        * \"median\": Border with the median value of the original border

    subsampling: str, optional
        The size of the returned nodemap. Defaults to "original".

    Raises
    ------
    ValueError
        If an argument is invalid.

    Returns
    -------
    ComponentTree
        The component tree.

    Example
    -------
    >>> import pylena as pln
    >>> t = pln.morpho.tos(img, root=(0, 0), padding="median")
    """
    if padding is not None and padding not in ("median"):
        raise ValueError(f"Invalid argument {padding}")
    subsampling_mode = ["original", "full", "full-no-border"]
    if subsampling not in subsampling_mode:
        raise ValueError(f"Invalid argument {subsampling}. Not in {subsampling_mode}")

    if padding == "median":
        img = add_border_median(img)
        if root:
            root = (root[0] + 1, root[1] + 1)

    if root:
        if not isinstance(root, tuple) or len(root) != 2:
            raise ValueError("Invalid root point")
        root = (int(root[0]), int(root[1]))
    else:
        root = (0, 0)

    tree, nodemap = cxx.tos(img, root)
    if subsampling == "original":
        nodemap = nodemap[::2, ::2]
    if padding is not None and subsampling != "full":
        nodemap = nodemap[1:-1, 1:-1]
    return ComponentTree(tree.parent, tree.values, nodemap)



# pylint: skip-file
# Setuptools
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.errors import OptionError, SetupError

# Conan
from conan.api.conan_api import ConanAPI
from conan.api.model import Remote
from conan.tools.env.environment import ProfileEnvironment
from conans.client.cache.cache import ClientCache

# Other
from tempfile import TemporaryDirectory
import pathlib
import subprocess
import pybind11
import os

from typing import Tuple

#-------------------------------------------------------------
# CMake Wrapper
#-------------------------------------------------------------

def _from_subprocess_output(subprocess_result : subprocess.Popen) -> Tuple[str, str]:
    return subprocess_result.stdout.decode("utf-8"), subprocess_result.stderr.decode("utf-8")

class CMake:
    def __init__(self, path=os.path.abspath(os.path.dirname(__file__)), variables={}, verbose=False):
        self.path = path
        self.variables = variables
        self.verbose = verbose

    def add_variable(self, name, value):
        if self.variables.get(name) is not None:
            raise ValueError(f"CMake variable {name} already set to {self.variable[name]}")
        self.variables[name] = value

    def configure(self):
        print("-- Configuring CMake")
        out = subprocess.run(["cmake", self.path] + [f"-D{variable}={value}" for (variable, value) in self.variables.items()], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if out.returncode != 0:
            stdout, stderr = _from_subprocess_output(out)
            raise SetupError(f"[ERROR]: No able to configure CMake\nSTDOUT:\n{stdout}\nSTDERR:\n{stderr}")
        print("-- CMake configured")

    def build(self, target=None):
        target_log = f" for target {target}" if target is not None else ""
        target_invoke = ["--target", target] if target is not None else []
        print(f"-- Building project{target_log}")
        out = subprocess.run(["cmake", "--build", "."] + target_invoke, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if out.returncode != 0:
            stdout, stderr = _from_subprocess_output(out)
            raise SetupError(f"[ERROR]: No able to build{target_log}\nSTDOUT:\n{stdout}\nSTDERR:\n{stderr}")
        print(f"-- Project built{target_log}")


#-------------------------------------------------------------
# Extension class
#-------------------------------------------------------------

class ConanCMakeExtension(Extension):
    def __init__(self, name, build_type="Release"):
        super().__init__(name, sources=[])
        if build_type not in ["Debug", "Release"]:
            raise OptionError(f"Extension build_type should be Debug or Release (Got {build_type}")
        self.build_type = build_type

class ConanCMakeBuildExtension(build_ext):
    def run(self):
        for ext in self.extensions:
            if not isinstance(ext, ConanCMakeExtension):
                raise SetupError("Extension is not ConanCMakeExtension")
            self._build(ext)

    def _configure_conan_profile(self, conan : ConanAPI, cache: ClientCache, ext):
        profile = conan.profiles.detect()
        if profile.settings["compiler"] == "Visual Studio":
            self.is_msvc = True
        profile.settings["compiler.cppstd"] = 20
        profile.settings["compiler.libcxx"] = "libstdc++11"
        profile.settings["build_type"] = f"{ext.build_type}"
        profile.process_settings(cache)
        print(profile.dumps())
        return profile

    def _conan_install(self, conan : ConanAPI, root, profile):
        LRDE_PUBLIC_URL = 'https://artifactory.lrde.epita.fr/artifactory/api/conan/lrde-public'
        remotes = conan.remotes.list(only_enabled=True)
        if not LRDE_PUBLIC_URL in (e.url for e in remotes):
            remotes.append(Remote("lrde-public", LRDE_PUBLIC_URL))
        path = conan.local.get_conanfile_path(root, str(self.absolute_build_temp), False)
        deps = conan.graph.load_graph_consumer(path, None, None, None, None, profile, profile, None, remotes, True)
        conan.graph.analyze_binaries(deps, ["missing", "onetbb/*", "fmt/*"], remotes=remotes)
        conan.install.install_binaries(deps, remotes)
        conan.install.install_consumer(deps, output_folder=str(self.absolute_build_temp))

    # Some code here come from https://stackoverflow.com/questions/42585210/extending-setuptools-extension-to-use-cmake-in-setup-py
    def _build(self, ext : ConanCMakeExtension):
        cwd = pathlib.Path().absolute()
        cache_folder = TemporaryDirectory()

        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)
        self.absolute_build_temp = build_temp.absolute()
        self.extdir = os.path.dirname(str(pathlib.Path(self.get_ext_fullpath(ext.name)).absolute()))

        os.chdir(self.build_temp)

        # Conan install
        conan = ConanAPI()
        cache = ClientCache(cache_folder.name)
        profile = self._configure_conan_profile(conan, cache, ext)
        self._conan_install(conan, str(cwd), profile)

        # CMake
        cmake_variables = self._make_cmake_variable(ext)
        cmake = CMake(cwd, cmake_variables)
        cmake.configure()
        if not self.dry_run:
            cmake.build()

        os.chdir(cwd)

    def _make_cmake_variable(self, ext : ConanCMakeExtension):
        cmake_variables = {
            "CMAKE_BUILD_TYPE": ext.build_type,
            "CMAKE_TOOLCHAIN_FILE": os.path.join(self.absolute_build_temp, "build", "conan_toolchain.cmake"),
            "pybind11_DIR": pybind11.get_cmake_dir(),
            "CMAKE_LIBRARY_OUTPUT_DIRECTORY": self.extdir,
            "CMAKE_BUILD_PARALLEL_LEVEL": 8
        }
        if "PYTHON_EXECUTABLE" in os.environ:
            cmake_variables["PYTHON_EXECUTABLE"] = os.environ["PYTHON_EXECUTABLE"]
        cmake_variables["CMAKE_POLICY_DEFAULT_CMP0091"] = "NEW"
        return cmake_variables

#-------------------------------------------------------------
# Setup
#-------------------------------------------------------------

setup(
    cmdclass={"build_ext": ConanCMakeBuildExtension},
    ext_modules=[ConanCMakeExtension("pylena/pylena_cxx")],
    platforms=["linux"],
    packages=find_packages(exclude=["tests", "doc"])
)



from ..pylena_cxx import morpho as cxx
from ..pylena_cxx.morpho import structuring_element_2d
from ..utils import check_numpy_array, check_type

from typing import Union

import numpy as np


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
@check_type(ind=1, wanted_types=[structuring_element_2d, np.ndarray])
def erosion(
    img: np.ndarray,
    se: Union[np.ndarray, structuring_element_2d],
    padding_value: int = None,
):
    """
    Performs an erosion by a structuring element.

    Given a structuring element :math:`B`, the erosion :math:`\\varepsilon(f)` of the input image
    :math:`f` is defined as

    .. math::

        \\varepsilon(f)(x) = \\bigwedge \\{f(y), y \\in B_x\\}

    Args
    ----
    img: 2-D array (dtype=uint8)
        The image to be processed
    se: structuring_element_2d or 2-D array
        The structuring element

    Return
    ------
    2-D array
        The resulting image

    Example
    -------

    >>> img = ... # Get an image
    >>> from pylena.morpho import make_structuring_element_2d, erosion
    >>> se = make_structring_element_2d("rect", 10, 10)
    >>> out = erosion(img, se)
    """
    if issubclass(type(se), np.ndarray):
        if se.ndim != 2:
            raise ValueError("Structuring element should be a 2D array")
        se = se.astype(bool)
        se = structuring_element_2d(se)
    return cxx.erosion(img, se, padding_value)


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
@check_type(ind=1, wanted_types=[structuring_element_2d, np.ndarray])
def dilation(
    img: np.ndarray,
    se: Union[np.ndarray, structuring_element_2d],
    padding_value: int = None,
):
    """
    Performs an dilation by a structuring element.

    Given a structuring element :math:`B`, the dilation :math:`\\delta(f)` of the input image
    :math:`f` is defined as

    .. math::

        \\delta(f)(x) = \\bigvee \\{f(y), y \\in B_x\\}

    Args
    ----
    img: 2-D array (dtype=uint8)
        The image to be processed
    se: structuring_element_2d or 2-D array
        The structuring element

    Return
    ------
    2-D array
        The resulting image

    Example
    -------

    >>> img = ... # Get an image
    >>> from pylena.morpho import make_structuring_element_2d, dilation
    >>> se = make_structring_element_2d("rect", 10, 10)
    >>> out = dilation(img, se)
    """
    if issubclass(type(se), np.ndarray):
        if se.ndim != 2:
            raise ValueError("Structuring element should be a 2D array")
        se = se.astype(bool)
        se = structuring_element_2d(se)
    if padding_value is not None:
        padding_value = int(padding_value)
    return cxx.dilation(img, se, padding_value)


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
@check_type(ind=1, wanted_types=[structuring_element_2d, np.ndarray])
def opening(
    img: np.ndarray,
    se: Union[np.ndarray, structuring_element_2d],
    padding_value: int = None,
):
    """
    Performs an opening by a structuring element.

    Given a structuring element :math:`B`, the dilation :math:`\\gamma(f)` of the input image
    :math:`f` is defined as

    .. math::

        \\gamma(f) = \\delta_{B}(\\varepsilon_{B}(f))

    Args
    ----
    img: 2-D array (dtype=uint8)
        The image to be processed
    se: structuring_element_2d or 2-D array
        The structuring element

    Return
    ------
    2-D array
        The resulting image

    Example
    -------

    >>> img = ... # Get an image
    >>> from pylena.morpho import make_structuring_element_2d, opening
    >>> se = make_structring_element_2d("rect", 10, 10)
    >>> out = opening(img, se)
    """
    if issubclass(type(se), np.ndarray):
        if se.ndim != 2:
            raise ValueError("Structuring element should be a 2D array")
        se = se.astype(bool)
        se = structuring_element_2d(se)
    if padding_value is not None:
        padding_value = int(padding_value)
    return cxx.opening(img, se, padding_value)


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
@check_type(ind=1, wanted_types=[structuring_element_2d, np.ndarray])
def closing(
    img: np.ndarray,
    se: Union[np.ndarray, structuring_element_2d],
    padding_value: int = None,
):
    """
    Performs an closing by a structuring element.

    Given a structuring element :math:`B`, the dilation :math:`\\gamma(f)` of the input image
    :math:`f` is defined as

    .. math::

        \\gamma(f) = \\varepsilon_{B}(\\delta_{B}(f))

    Args
    ----
    img: 2-D array (dtype=uint8)
        The image to be processed
    se: structuring_element_2d or 2-D array
        The structuring element

    Return
    ------
    2-D array
        The resulting image

    Example
    -------

    >>> img = ... # Get an image
    >>> from pylena.morpho import make_structuring_element_2d, closing
    >>> se = make_structring_element_2d("rect", 10, 10)
    >>> out = closing(img, se)
    """
    if issubclass(type(se), np.ndarray):
        if se.ndim != 2:
            raise ValueError("Structuring element should be a 2D array")
        se = se.astype(bool)
        se = structuring_element_2d(se)
    if padding_value is not None:
        padding_value = int(padding_value)
    return cxx.closing(img, se, padding_value)


@check_numpy_array(ind=0, dtypes=[np.uint8], ndims=[2])
@check_type(ind=1, wanted_types=[structuring_element_2d, np.ndarray])
def gradient(
    img: np.ndarray,
    se: Union[np.ndarray, structuring_element_2d],
    padding_value: int = None,
):
    """
    Performs a morphological gradient, also called **Beucher** gradient.

    Given a structuring element :math:`B` and an image :math:`f`, the morphological gradient is defined as

    .. math::

        \\rho_B = \\delta_B - \\varepsilon_B

    Args
    ----
    img: 2-D array (dtype=uint8)
        The image to be processed
    se: structuring_element_2d or 2-D array
        The structuring element

    Return
    ------
    2-D array
        The resulting image

    Example
    -------

    >>> img = ... # Get an image
    >>> from pylena.morpho import make_structuring_element_2d, gradient
    >>> se = make_structring_element_2d("rect", 10, 10)
    >>> out = gradient(img, se)
    """
    if issubclass(type(se), np.ndarray):
        if se.ndim != 2:
            raise ValueError("Structuring element should be a 2D array")
        se = se.astype(bool)
        se = structuring_element_2d(se)
    if padding_value is not None:
        padding_value = int(padding_value)  # coverage: disable
    return cxx.gradient(img, se, padding_value)



import re
import pytest

from pylena import __version__

# From https://www.python.org/dev/peps/pep-0440/ with adjustment to pytest
def test_version():
    assert (
        re.match(
            r"^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$",
            __version__,
        )
        is not None
    )



# pylint: skip-file
from conan import ConanFile
from conan.tools.build import check_min_cppstd
from conan.tools.cmake import CMakeToolchain, CMakeDeps, CMake

class PyleneNumpyConan(ConanFile):
    name = "pylene-numpy"
    version = "head"
    settings = "os", "compiler", "build_type", "arch"
    options = {"shared": [True, False], "fPIC": [True, False]}
    default_options = {"shared": False, "fPIC": True, "pylene/*:fPIC": True}
    exports_sources = "CMakeLists.txt", "pylene-numpy/*", "LICENCE"

    def requirements(self):
        self.requires("pylene/head@lrde/unstable", transitive_libs=True, transitive_headers=True)
        self.requires("pybind11/2.9.2", transitive_headers=True)

    def validate(self):
        check_min_cppstd(self, 20)

    def configure(self):
        if self.options.shared:
            self.options.fPIC = True
        if self.settings.os != "Windows" and not self.options.fPIC:
            self.output.error("pylene-numpy is intended to be linked to python module and should be compiled with fPIC")

    def generate(self):
        deps = CMakeDeps(self)
        deps.generate()
        tc = CMakeToolchain(self)
        tc.generate()

    def layout(self):
        self.folders.source = "."
        self.folders.build = "build"
        self.folders.generators = "build"

        self.cpp.source.includedirs = ["pylene-numpy/include"]
        self.cpp.build.libdirs = ["pylene-numpy"]
        self.cpp.package.libs = ["lib"]
        self.cpp.package.includedirs = ["include"]

    def build(self):
        variables = {"BUILD_PYLENA": "OFF"}
        cmake = CMake(self)
        cmake.configure(variables)
        cmake.build()
        cmake.install()

    def package_info(self):
        self.cpp_info.set_property("cmake_target_name", "pylene-numpy::pylene-numpy")

        # Core pylene numpy
        self.cpp_info.requires = ["pylene::core", "pybind11::pybind11"]
        self.cpp_info.libs = ["pylene-numpy"]